import threading
from typing import Optional, Callable, Dict, List, Any


class SharedEventSystem:
    """
    This implementation uses a singleton pattern and maintains one global dictionary
    for all objects to store their subscriptions and run their event system.
    """

    _instance: Optional["SharedEventSystem"] = None
    _lock = threading.Lock()
    _subscribers: Dict[str, List[Callable]]

    def __init__(self):
        if not SharedEventSystem._instance:
            SharedEventSystem._instance = self

    @classmethod
    def initialize(cls):
        if not cls._instance:
            with cls._lock:
                cls._instance = cls()
                cls._subscribers = {}

    @classmethod
    def subscribe(cls, event_type: str, fn: Callable):
        if not cls._instance:
            cls.initialize()
        if fn is not None:
            if event_type not in cls._subscribers:
                cls._subscribers[event_type] = []
            cls._subscribers[event_type].append(fn)

    @classmethod
    def post(cls, event_type: str, event_data: Any):
        """
        Post an event to all subscribers of a specified event type.

        Raises:
            Exception: If an instance of EventsSystem has not been initialized.
            Note: This happens, if post() is being called before subscribe() has not
            been called at least once, which leads to the instance of EventsSystem being
            None.
        """
        if cls._instance is None:
            raise Exception(
                "At least one subscription has to be registered before posting events."
            )

        if event_type in cls._subscribers:
            for fn in cls._subscribers[event_type]:
                fn(event_data)



from event_systems.internal.event_system import InternalEventSystem
from tests.helpers.dummy_emitter import DummyEmitter
from tests.helpers.dummy_subscriber import (
    DummySubscriber_zero,
    DummySubscriber_one,
    DummySubscriber_two,
)


def test_events_system_subscribe_zero_subscription_results_in_zero_entries():
    # given
    es = InternalEventSystem()
    _ = DummyEmitter(es)

    # when
    _ = DummySubscriber_zero(es)

    # then
    assert len(es._subscribers) == 0


def test_events_system_subscribe_one_subscription_results_in_one_entry():
    # given
    es = InternalEventSystem()
    _ = DummyEmitter(es)

    # when
    _ = DummySubscriber_one(es)

    # then
    assert len(es._subscribers) == 1


def test_events_system_subscribe_two_subscriptions_results_in_two_entries():
    # given
    es = InternalEventSystem()
    _ = DummyEmitter(es)

    # when
    _ = DummySubscriber_two(es)

    # then
    assert len(es._subscribers) == 2


def test_events_system_post_zero_subscriptions_results_in_zero_events_handled(capsys):
    # given
    es = InternalEventSystem()
    emitter = DummyEmitter(es)
    _ = DummySubscriber_zero(es)

    # when
    emitter.emit_event()

    # then
    out, err = capsys.readouterr()
    expected = ""
    assert expected == out


def test_events_system_post_one_subscription_results_in_one_event_handled(capsys):
    # given
    es = InternalEventSystem()
    emitter = DummyEmitter(es)
    _ = DummySubscriber_one(es)

    # when
    emitter.emit_event()

    # then
    out, err = capsys.readouterr()
    expected = "DummySubscriber_one - event handeled\n"
    assert expected == out


def test_events_system_post_two_subscriptions_results_in_two_events_handled(capsys):
    # given
    es = InternalEventSystem()
    emitter = DummyEmitter(es)
    _ = DummySubscriber_two(es)

    # when
    emitter.emit_event()
    emitter.emit_another_event()

    # then
    out, err = capsys.readouterr()
    expected_1 = "DummySubscriber_two - event handeled\n"
    expected_2 = "DummySubscriber_two - another event handeled\n"
    assert out == expected_1 + expected_2


def test_post_with_event_data_one_subscription_results_in_one_event_handled_with_data(
    capsys,
):
    # given
    es = InternalEventSystem()
    emitter = DummyEmitter(es)
    _ = DummySubscriber_one(es)

    # when
    emitter.emit_event("some event data")

    # then
    out, err = capsys.readouterr()
    expected = "DummySubscriber_one - event handeled some event data\n"
    assert out == expected


def test_events_system_initialize_two_objects_results_in_seperate_identities():
    # given
    event_system_1 = InternalEventSystem()
    event_system_2 = InternalEventSystem()

    # when
    object_id_1 = id(event_system_1)
    object_id_2 = id(event_system_2)

    assert object_id_1 != object_id_2


def test_subscribe_to_two_instances_results_in_each_having_one_subscription():
    # given
    es_1 = InternalEventSystem()
    es_2 = InternalEventSystem()

    # when
    _ = DummySubscriber_one(es_1)
    _ = DummySubscriber_one(es_2)

    # then
    assert len(es_1._subscribers) == 1
    assert len(es_2._subscribers) == 1



import threading
from typing import Callable, Dict, List, Any


class InternalEventSystem:
    """
    This implementation uses a factory pattern and allows individual objects to have
    their own event system, passed as a variable to each of their child objects.
    """

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._subscribers: Dict[str, List[Callable]] = {}

    def subscribe(self, event_type: str, fn: Callable):
        with self._lock:
            if fn is not None:
                if event_type not in self._subscribers:
                    self._subscribers[event_type] = []
                self._subscribers[event_type].append(fn)

    def post(self, event_type: str, event_data: Any):
        if event_type in self._subscribers:
            for fn in self._subscribers[event_type]:
                fn(event_data)

    def get_subscribers(self) -> Dict[str, List[Callable]]:
        return self._subscribers
